#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
@File: datasplit.py
@Time: 2019-01-11
@Author: ternencewang2015@outlook.com
@Direct: generate txt file in Main
"""

import os
import glob
import random


class DataSplit():
    def __init__(self, root_dir):
        self.root_dir = root_dir

        self.txt_dir = os.path.join(self.root_dir, 'ImageSets/Main')
        self.xmls_dir = os.path.join(self.root_dir, 'Annotations')
        self.images_dir = os.path.join(self.root_dir, 'JPEGImages')
        assert os.path.exists(self.xmls_dir), self.xmls_dir
        assert os.path.exists(self.images_dir), self.images_dir

        if not os.path.exists(self.txt_dir):
            os.makedirs(self.txt_dir)

        self.useful_name_list = self.match_xml_and_jpg()

    def match_xml_and_jpg(self, xmls_dir=None, images_dir=None):
        if xmls_dir == None:
            xmls_dir = self.xmls_dir
        if images_dir == None:
            images_dir = self.images_dir

        xmls_list = sorted(glob.glob(os.path.join(xmls_dir, '*.xml')))
        useful_name_list = []
        for xml_path in xmls_list:
            name = os.path.basename(xml_path).split('.')[0]
            image_path = os.path.join(images_dir, '{}.jpg'.format(name))
            if os.path.exists(image_path):
                useful_name_list.append(name)
                
        return useful_name_list

    def prefix_grouping(self, prefix_list, name_list=None):
        if name_list is None:
            name_list = self.useful_name_list

        groups = {}
        for prefix in prefix_list:
            groups[prefix] = []

        for name in name_list:
            for prefix in prefix_list:
                if name.startswith(prefix):
                    groups[prefix].append(name)

        return groups

    def split_by_rate(self, test_rate, val_rate=0.0, name_list=None, shuffle=False):
        if name_list is None:
            name_list = self.useful_name_list

        assert test_rate < 1, 'Error: test_rate {} not in range.'.format(test_rate)
        assert len(name_list) > 2, 'Error: name_list length is needed more than 2.'

        if shuffle:
            random.shuffle(name_list)

        test_number = int(test_rate * len(name_list))
        test_number = test_number if test_number > 0 else 1

        if val_rate > 0:
            val_number = int(val_rate * len(name_list))
            val_number = val_number if val_number > 0 else 1
        else:
            val_number = 0

        train_number = len(name_list) - test_number - val_number
        assert train_number > 0, 'Error: train_number is needed more than 0.'
        
        train_list = name_list[0:train_number]
        test_list = name_list[train_number:train_number+test_number]
        if val_number > 0:
            val_list = name_list[-val_number:]
        else:
            val_list = []
        return {'train': train_list, 'val': val_list, 'test': test_list}

    def split_group_by_rate(self, groups, test_rate, val_rate=0.0, shuffle=False):
        """
        Arguments:
            groups: dict like {prefix: [name, ]}, grouped name list.
            test_rate: float, the test rate of all data.
            val_rate: float, the val rate of all data, default is 0.0.
            shuffle: bool, default is False.
        Returns:
            result: dict like {'train': [], 'val': [], 'test': []}.
        """
        train_list, val_list, test_list = [], [], []

        for prefix in groups:
            if not groups[prefix]:
                continue
            split_result = self.split_by_rate(test_rate, val_rate, name_list=groups[prefix], shuffle=shuffle)
            train_list = train_list + split_result['train']
            test_list = test_list + split_result['test']
            val_list = val_list + split_result['val']

        return {'train': train_list, 'val': val_list, 'test': test_list}

    def save(self, split_name_dic, save_dir=None):
        if save_dir is None:
            save_dir = self.txt_dir

        if 'trainval' not in split_name_dic:
            split_name_dic['trainval'] = split_name_dic['train'] + split_name_dic['val']

        for key, val in split_name_dic.items():
            with open(os.path.join(save_dir, key+'.txt'), 'w') as f:
                f.write('\n'.join(val))

def test():
    root_dir = '/home/wangtf/ShareDataset/dataset/RebarDataset/rebar-test/VOC2007'
    prefix_list = ['12', '14', '16', '18', '20', '22', '25', '32']
    test_rate = 0.2
    val_rate = 0.2

    spliter = DataSplit(root_dir)
    groups = spliter.prefix_grouping(prefix_list)
    split_result = spliter.split_group_by_rate(groups, test_rate, val_rate=val_rate)
    spliter.save(split_result)


if __name__ == '__main__':
    test()




#!/usr/bin/evn python3
# -*- coding:utf-8 -*-
"""
@File: decodeXml.py
@Author: ternencewang
@Desc: decode xml file of Pascal Voc annotation

"""
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


class XmlReader():
    """Decode the xml file"""
    def __init__(self, xml_file_path, debug=False):
        """
        Arguments:
            xml_file_path: str, the path of a xml file
        """
        self.debug = debug
        self.tree = None
        self.root = None
        self.rewrite = False
        self.file_path = xml_file_path
        self.get_root()

    def get_root(self):
        """get root node of the xml
        """
        self.tree = ET.parse(self.file_path)  #打开xml文档
        self.root = self.tree.getroot()         #获得root节点

    def get_file_name(self):
        """Get file name node information in xml"""
        print("*"*10)
        filename = self.root.find('filename').text
        filename = filename[:-4]
        print(filename)

    def get_width_and_height(self):
        """Get width and height of image in xml"""
        for size in self.root.findall('size'): # 找到root节点下的size节点
            width = size.find('width').text   # 子节点下节点width的值
            height = size.find('height').text   # 子节点下节点height的值
            print(width, height)

    def get_object_list(self):
        """Get all object name and number"""
        obj_list = {}
        for object in self.root.findall('object'):
            name = self.get_object_name(object)
            if name not in obj_list:
                obj_list[name] = 1
            else:
                obj_list[name] += 1
        return obj_list

    def get_object_name(self, object):
        """Get object's name
        Arguments:
            object: object, the the of xml node
        Return:
            name: str, the name of object
        """
        name = object.find('name').text
        # print('Object name is :{}'.format(name))
        return name
    
    def get_object_pose(self, object):
        pose = object.find('pose').text
        return pose

    def get_object_truncated(self, object):
        truncated = object.find('truncated').text
        return truncated

    def get_object_difficult(self, object):
        difficult = object.find('difficult').text
        return difficult

    def set_object_name(self, old_name, new_name):
        """replace old_name to new_name for all object nodes
        Arguments:
            old_name: str, the old class name
            new_name: str, the new class name
        """
        if self.debug:
            print('old_name: {}, new_name: {}'.format(old_name, new_name))
        for object in self.root.findall('object'):
            name = self.get_object_name(object)
            if name == old_name:
                object.find('name').text = new_name
                self.rewrite = True
                if self.debug:
                    print('will replace label name.')
        if self.rewrite is True:
            self.rewrite_xml()
            print('Rerite xml: {}'.format(self.file_path))
        
    def rewrite_xml(self, save_path=None):
        """Save xml file"""
        if save_path is None:
            save_path = self.file_path
        self.tree.write(save_path)
    
    def get_object_bndbox(self, object):
        """Get bbox in object node
        Arguments:
            object: object, the type of xml node
        Return:
            a string list of [xmin, ymin, xmax, ymax]
        """
        bndbox = object.find('bndbox')      #子节点下属性bndbox的值
        xmin = bndbox.find('xmin').text
        ymin = bndbox.find('ymin').text
        xmax = bndbox.find('xmax').text
        ymax = bndbox.find('ymax').text
        return [xmin, ymin, xmax, ymax]

    def set_object_bndbox(self, scale, pad=[0, 0], save_path=None):
        """reset object bndbox using sacle and pad
        Arguments:
            scale: float, the scale chang between original image and new image.
            pad: list of two int, top pad width and left pad width, the down pad is same as the top, the right pad is same as the left.
            save_path: str, the new xml save path, default is rewrite original file.
        """
        for obj in self.root.findall('object'):
            bndbox = obj.find('bndbox')
            bndbox.find('xmin').text = str(round(int(bndbox.find('xmin').text) * scale) + pad[1])  # add left pad
            bndbox.find('ymin').text = str(round(int(bndbox.find('ymin').text) * scale) + pad[0])  # add top pad
            bndbox.find('xmax').text = str(round(int(bndbox.find('xmax').text) * scale) + pad[1])
            bndbox.find('ymax').text = str(round(int(bndbox.find('ymax').text) * scale) + pad[0])
        self.rewrite_xml(save_path)

    def get_all_object_bndbox(self):
        bboxes = []
        for obj in self.root.findall('object'):
            bbox = self.get_object_bndbox(obj)
            bbox = [int(x) for x in bbox]
            bboxes.append(bbox)
    
        return bboxes

    def get_all_object(self):
        objects = []
        for obj in self.root.findall('object'):
            bbox = self.get_object_bndbox(obj)
            bbox = [int(x) for x in bbox]
            name = self.get_object_name(obj)
            pose = self.get_object_pose(obj)
            truncated = self.get_object_truncated(obj)
            difficult = self.get_object_difficult(obj)

            objects.append({'name': name, 'bbox': bbox, 'pose': pose, 'truncated': truncated, 'difficult': difficult})

        return objects

    def save_element_info(self, save_dict, tag, text):
        """Save tag and text to save_dict.
        if tag not in save_dict, it will return like save_dict[tag] = text,
        otherwise it will like save_dict[tag] = [text, ]
        Arguments:
            save_dict: dict, to save.
            tag: str, the key to save.
            text: str, the value to save.
        """
        if tag not in save_dict:
            save_dict[tag] = text
        else:
            if not isinstance(save_dict[tag], list):
                save_dict[tag] = [save_dict[tag]]
            save_dict[tag].append(text)

    def parse_element(self, element, save_dict=None):
        """Parse all information in element and save to save_dict.
        Arguments:
            element: element, an element type node in xml tree.
            save_dict: dict, save result.
        Returns:
            save_dict: dict, like {'path': './', 'segmented': '0', 'object': [{}]}.
        """
        if save_dict is None:
            save_dict = {}

        for child in element:
            if len(child) == 0:
                self.save_element_info(save_dict, child.tag, child.text)
            else:
                self.save_element_info(save_dict, child.tag, self.parse_element(child))

        return save_dict

    def parse_annotations(self):
        """ Parse all annotations under the xml_root.
        """
        xml_root = self.root
        annotations = self.parse_element(xml_root)
        return annotations

    def load(self):
        return self.parse_annotations()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-


import os


class CropData():
    def __init__(self, root_dir, save_root_dir):
        self.root_dir = root_dir
        self.save_root_dir = save_root_dir
        

    def crop(image_path, xml_path, )


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@File: image_annotation_split.py
@Author: ternencewang
@Time: 2019-01-14
@Desc: split image and annotation.
"""

import os
import glob
import numpy as np
import cv2

from .xmlreader import XmlReader
from .xmlwriter import XmlWriter
from .iou import bb_intersection_over_union as iou


class SplitImageAnnotation():
    """Crop subimage and subannotation in data and generate an new dataset."""
    def __init__(self, jpg_dir, xml_dir, save_jpg_dir, save_xml_dir, image_min_side=1000, image_max_side=1333):
        self.jpg_dir = jpg_dir
        self.xml_dir = xml_dir
        self.save_jpg_dir = save_jpg_dir
        self.save_xml_dir = save_xml_dir

        self.image_min_side = image_min_side
        self.image_max_side = image_max_side

    def load_images_path(self):
        images_path = glob.glob(os.path.join(self.jpg_dir, '*.jpg'))
        return images_path

    def load_xmls_path(self):
        xmls_path = glob.glob(os.path.join(self.xml_dir, '*.xml'))
        return xmls_path

    def match_jpg_xml(self):
        images_path = self.load_images_path()
        xmls_path = self.load_xmls_path()
        images_name = [os.path.basename(path).split('.')[0] for path in images_path]
        xmls_name = [os.path.basename(path).split('.')[0] for path in xmls_path]

        matched_name = [name for name in xmls_name if name in images_name]

        matched_images_path = [os.path.join(self.jpg_dir, name+'.jpg') for name in matched_name]
        matched_xmls_path = [os.path.join(self.xml_dir, name+'.xml') for name in matched_name]
        return matched_images_path, matched_xmls_path

    def get_split_bboxes(self, width, height, cover_thresh=0.2):
        bboxes = []
        if width >= height:
            x_stride = self.image_max_side
            y_stride = self.image_min_side
        else:
            x_stride = self.image_min_side
            y_stride = self.image_max_side

        def add_row_crop(start_y):
            start_x = 0
            while start_x + x_stride < width:
                bboxes.append([start_x, start_y, start_x+x_stride, start_y+y_stride])
                start_x += int(x_stride * (1-cover_thresh))
            bboxes.append([width-x_stride, start_y, width, start_y+y_stride])

        start_y = 0
        while start_y + y_stride < height:
            add_row_crop(start_y)
            start_y += int(y_stride * (1-cover_thresh))
        add_row_crop(height-y_stride)
        return bboxes

    def split_image(self, image, split_bboxes):
        """Split an image to some subimages.
        Arguments:
            image: ndarray, image data.
            split_bboxes: list, like[[xmin, ymin, xmax, ymax], ]
        Returns:
            images: list, all subimages.
            left_top_strides: list, (left, top) point strides.
            right_down_strides: list, (right, down) point strides.
        """
        subimages = []
        for bbox in split_bboxes:
            subimages.append(image[bbox[1]:bbox[3], bbox[0]:bbox[2]])
        return subimages

    def split_annotations(self, xml_info, split_bboxes, iou_thresh=0.7):
        """Using split_bboxes to split an xml file.
        Arguments:
            xml_info: dict, all info about a xml.
            split_bboxes: list, like [[xmin, ymin, xmax, ymax], ]
        Returns:
            subannotations: list, like [xml_info, ]
        """
        subannotations = []
        for bbox in split_bboxes:
            xmin, ymin, xmax, ymax = bbox

            # init sub xml info
            sub_xml_info = {
                'path': xml_info['path'],
                'filename': xml_info['filename'],
                'folder': xml_info['folder'],
                'width': int(xmax)-int(xmin),
                'height': int(ymax)-int(ymin),
                'depth': xml_info['depth'],
                'database': xml_info['database'],
                'segmented': xml_info['segmented'],
                'objects': []
            }

            for bbox_info in xml_info['objects']:
                ob_xmin, ob_ymin, ob_xmax, ob_ymax = bbox_info['xmin'], bbox_info['ymin'], bbox_info['xmax'], bbox_info['ymax'], 
                if iou([ob_xmin, ob_ymin, ob_xmax, ob_ymax], [xmin, ymin, xmax, ymax]) > iou_thresh:
                    sub_xml_info['objects'].append({
                        'name': bbox_info['name'],
                        'xmin': max(ob_xmin - xmin, 1),
                        'ymin': max(ob_ymin - ymin, 1),
                        'xmax': min(ob_xmax - xmin, xmax-xmin-1),
                        'ymax': min(ob_ymax - ymin, ymax-ymin-1),
                        'pose': bbox_info['pose'],
                        'truncated': bbox_info['truncated'],
                        'difficult': bbox_info['difficult'],
                        })
            subannotations.append(sub_xml_info)

        return subannotations

    def get_annotations(self, xml_path):
        xml_reader = XmlReader(xml_path)
        annotations = xml_reader.get_all_object()
        return annotations

    def split_dir(self):
        images_path, xmls_path = self.match_jpg_xml()
        print('images: {}, xmls: {}'.format(len(images_path), len(xmls_path)))

        for index, (jpg_path, xml_path) in enumerate(zip(images_path, xmls_path)):
            print('{}: split image: {}\t\t\t'.format(index, os.path.basename(jpg_path)), end='\r')
            images, xml_writers = self.split_image_annotations(jpg_path, xml_path)
            for i, (image, xml_writer) in enumerate(zip(images, xml_writers)):
                save_xml_path = os.path.join(self.save_xml_dir, os.path.basename(xml_path).replace('.xml', '_{:0>2d}.xml'.format(i)))
                save_jpg_path = os.path.join(self.save_jpg_dir, os.path.basename(jpg_path).replace('.jpg', '_{:0>2d}.jpg'.format(i)))
                cv2.imwrite(save_jpg_path, image)
                xml_writer.save(save_xml_path, image_parameters={'path': save_jpg_path})

    def split_image_annotations(self, jpg_path, xml_path, cover_thresh=0.2, iou_thresh=1.0, database='Unknown', segmented=0):
        """Split an image and it's annotation file.
        Arguments:
            jpg_path: str, image path.
            xml_path: str, xml path.
            cover_thresh: float, cover rate about each subimage, default=0.2.
            iou_thresh: float, filter annotations which one's iou is smaller than thresh, default=1.0.
            database: str, save xml database name.
            segmented: bool, save xml segmented name.
        """
        image = cv2.imread(jpg_path)
        split_bboxes = self.get_split_bboxes(width=image.shape[1], height=image.shape[0], cover_thresh=cover_thresh)
        subimages = self.split_image(image, split_bboxes)

        annotations = self.get_annotations(xml_path)
        subannotations = self.split_annotations(annotations, split_bboxes)
        
        return subimages, subannotations

    def copy_split_data_name(self):
        root_dir = os.path.join(self.xml_dir, '../ImageSets/Main')
        train_path = os.path.join(root_dir, 'train.txt')
        val_path = os.path.join(root_dir, 'val.txt')
        test_path = os.path.join(root_dir, 'test.txt')

        assert os.path.exists(train_path), train_path
        assert os.path.exists(val_path), val_path
        assert os.path.exists(test_path), test_path

        save_path = os.path.join(self.save_xml_dir, '../ImageSets/Main')
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        
        for txt_path in [train_path, val_path, test_path]:
            with open(txt_path) as file:
                name_list = file.readlines()
                new_path_list = []
                for name in name_list:
                    new_path_list += glob.glob(os.path.join(self.save_xml_dir, name.strip()+'*.xml'))
                new_name_list = [os.path.basename(name).split('.')[0] for name in new_path_list]
                with open(os.path.join(save_path, os.path.basename(txt_path)), 'w') as f:
                    f.write('\n'.join(new_name_list))


def test():
    root_dir = '/home/wangtf/ShareDataset/dataset/RebarDataset/VOCdevkit_rebar_v9_0-12-14-16-18-20-22-25-32'
    jpg_dir = os.path.join(root_dir, 'VOC2007/JPEGImages')
    xml_dir = os.path.join(root_dir, 'VOC2007/Annotations')
    assert os.path.exists(xml_dir), xml_dir

    save_root_dir = '{}_split_test/VOC2007'.format(root_dir)
    save_jpg_dir = os.path.join(save_root_dir, 'JPEGImages')
    save_xml_dir = os.path.join(save_root_dir, 'Annotations')
    if not os.path.exists(save_jpg_dir):
        os.makedirs(save_jpg_dir)
    if not os.path.exists(save_xml_dir):
        os.makedirs(save_xml_dir)
    
    spliter = SplitImageAnnotation(jpg_dir, xml_dir, save_jpg_dir, save_xml_dir)
    spliter.split_dir()
    spliter.copy_split_data_name()


if __name__ == '__main__':
    test()



from ._version import version
from .xmlreader import XmlReader
from .xmlwriter import XmlWriter
from .resize import DatasetResize
from .datasplit import DataSplit
from .image_annotation_split import SplitImageAnnotation

name = 'pascal_voc_tools'
__version__ = version




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@File: xmlwriter.py
@Time: 2019-01-14
@Author: ternencewang
@Direc: write a xml file about pascal voc annotation.
"""
import os
from jinja2 import Environment, PackageLoader


class XmlWriter():
    """Write a xml file about pascal voc annotation."""
    def __init__(self, path='', width=0, height=0, depth=3, database='Unknown', segmented=0):
        """Generate a xml file
        Arguments:
            path: str, arg in xml about image.
            width: int or str, image width.
            height: int or str, image height.
            depth: int or str, image channle.
            database: str, default='Unknown'.
            segmented: int or str, default=0.
        """
        environment = Environment(loader=PackageLoader('pascal_voc_tools', 'templates'), keep_trailing_newline=True)
        self.annotation_template = environment.get_template('annotation.xml')

        abspath = os.path.abspath(path)

        self.template_parameters = {
            'path': abspath,
            'filename': os.path.basename(abspath),
            'folder': os.path.basename(os.path.dirname(abspath)),
            'width': width,
            'height': height,
            'depth': depth,
            'database': database,
            'segmented': segmented,
            'object': []
        }

    def add_object(self, name, xmin, ymin, xmax, ymax, pose='Unspecified', truncated=0, difficult=0):
        """add an object info
        Arguments:
            name: str, class name.
            xmin: int, left.
            ymin: int, top.
            xmax: int, right.
            ymax: int, bottom.
            pose: str, default is 'Unspecified'.
            truncated: str, default is 0.
            difficult: int, default is 0.
        """
        self.template_parameters['object'].append({
            'name': name,
            'pose': pose,
            'truncated': truncated,
            'difficult': difficult,
            'bndbox': {
                'xmin': xmin,
                'ymin': ymin,
                'xmax': xmax,
                'ymax': ymax,
            },
        })

    def save(self, annotation_path, image_parameters=None):
        """Write a xml file to save info.
        Arguments:
            annotation_path: str, the path of xml to save.
        """
        if image_parameters is not None:
            for k, v in image_parameters.items():
                self.template_parameters[k] = image_parameters[k]

        with open(annotation_path, 'w') as xml_file:
            content = self.annotation_template.render(**self.template_parameters)
            xml_file.write(content)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import numpy as np
import random
import warnings

import keras

from ..utils.anchors import (
    anchor_targets_bbox,
    anchors_for_shape,
    guess_shapes
)
from ..utils.config import parse_anchor_parameters
from ..utils.image import (
    TransformParameters,
    adjust_transform_for_image,
    apply_transform,
    preprocess_image,
    resize_image,
)
from ..utils.transform import transform_aabb


class Generator(keras.utils.Sequence):
    """ Abstract generator class.
    """

    def __init__(
        self,
        batch_size,
        image_min_side,
        image_max_side,
        transform_generator,# = None,
        group_method='ratio',  # one of 'none', 'random', 'ratio'
        shuffle_groups=True,
        transform_parameters=None,
        compute_anchor_targets=anchor_targets_bbox,
        compute_shapes=guess_shapes,
        preprocess_image=preprocess_image,
        config=None,
        preprocess_image_mode='caffe'
    ):
        """ Initialize Generator object.

        Args
            transform_generator    : A generator used to randomly transform images and annotations.
            batch_size             : The size of the batches to generate.
            group_method           : Determines how images are grouped together (defaults to 'ratio', one of ('none', 'random', 'ratio')).
            shuffle_groups         : If True, shuffles the groups each epoch.
            image_min_side         : After resizing the minimum side of an image is equal to image_min_side.
            image_max_side         : If after resizing the maximum side is larger than image_max_side, scales down further so that the max side is equal to image_max_side.
            transform_parameters   : The transform parameters used for data augmentation.
            compute_anchor_targets : Function handler for computing the targets of anchors for an image and its annotations.
            compute_shapes         : Function handler for computing the shapes of the pyramid for a given input.
            preprocess_image       : Function handler for preprocessing an image (scaling / normalizing) for passing through a network.
        """
        self.transform_generator    = transform_generator
        self.batch_size             = int(batch_size)
        self.group_method           = group_method
        self.shuffle_groups         = shuffle_groups
        self.image_min_side         = image_min_side
        self.image_max_side         = image_max_side
        self.transform_parameters   = transform_parameters or TransformParameters()
        self.compute_anchor_targets = compute_anchor_targets
        self.compute_shapes         = compute_shapes
        self.preprocess_image       = preprocess_image
        self.config                 = config
        self.preprocess_image_mode  = preprocess_image_mode

        # Define groups
        self.group_images()

        # Shuffle when initializing
        if self.shuffle_groups:
            self.on_epoch_end()

    def on_epoch_end(self):
        random.shuffle(self.groups)

    def size(self):
        """ Size of the dataset.
        """
        raise NotImplementedError('size method not implemented')

    def num_classes(self):
        """ Number of classes in the dataset.
        """
        raise NotImplementedError('num_classes method not implemented')

    def has_label(self, label):
        """ Returns True if label is a known label.
        """
        raise NotImplementedError('has_label method not implemented')

    def has_name(self, name):
        """ Returns True if name is a known class.
        """
        raise NotImplementedError('has_name method not implemented')

    def name_to_label(self, name):
        """ Map name to label.
        """
        raise NotImplementedError('name_to_label method not implemented')

    def label_to_name(self, label):
        """ Map label to name.
        """
        raise NotImplementedError('label_to_name method not implemented')

    def image_aspect_ratio(self, image_index):
        """ Compute the aspect ratio for an image with image_index.
        """
        raise NotImplementedError('image_aspect_ratio method not implemented')

    def load_image(self, image_index):
        """ Load an image at the image_index.
        """
        raise NotImplementedError('load_image method not implemented')

    def load_annotations(self, image_index):
        """ Load annotations for an image_index.
        """
        raise NotImplementedError('load_annotations method not implemented')

    def load_annotations_group(self, group):
        """ Load annotations for all images in group.
        """
        annotations_group = [self.load_annotations(image_index) for image_index in group]
        for annotations in annotations_group:
            assert(isinstance(annotations, dict)), '\'load_annotations\' should return a list of dictionaries, received: {}'.format(type(annotations))
            assert('labels' in annotations), '\'load_annotations\' should return a list of dictionaries that contain \'labels\' and \'bboxes\'.'
            assert('bboxes' in annotations), '\'load_annotations\' should return a list of dictionaries that contain \'labels\' and \'bboxes\'.'

        return annotations_group

    def filter_annotations(self, image_group, annotations_group, group):
        """ Filter annotations by removing those that are outside of the image bounds or whose width/height < 0.
        """
        # test all annotations
        for index, (image, annotations) in enumerate(zip(image_group, annotations_group)):
            # test x2 < x1 | y2 < y1 | x1 < 0 | y1 < 0 | x2 <= 0 | y2 <= 0 | x2 >= image.shape[1] | y2 >= image.shape[0]
            invalid_indices = np.where(
                (annotations['bboxes'][:, 2] <= annotations['bboxes'][:, 0]) |
                (annotations['bboxes'][:, 3] <= annotations['bboxes'][:, 1]) |
                (annotations['bboxes'][:, 0] < 0) |
                (annotations['bboxes'][:, 1] < 0) |
                (annotations['bboxes'][:, 2] > image.shape[1]) |
                (annotations['bboxes'][:, 3] > image.shape[0])
            )[0]

            # delete invalid indices
            if len(invalid_indices):
                warnings.warn('Image with id {} (shape {}) contains the following invalid boxes: {}.'.format(
                    group[index],
                    image.shape,
                    annotations['bboxes'][invalid_indices, :]
                ))
                for k in annotations_group[index].keys():
                    annotations_group[index][k] = np.delete(annotations[k], invalid_indices, axis=0)

        return image_group, annotations_group

    def load_image_group(self, group):
        """ Load images for all images in a group.
        """
        return [self.load_image(image_index) for image_index in group]

    def random_transform_group_entry(self, image, annotations, transform=None):
        """ Randomly transforms image and annotation.
        """

        # wangtf: Try, move resize func from preprocess to here, maybe better
        # resize image
        image, image_scale = self.resize_image(image, min_side=self.image_min_side, max_side=self.image_max_side)

        # apply resizing to annotations too
        annotations['bboxes'] *= image_scale

        # randomly transform both image and annotations
        if transform is not None or self.transform_generator:
            if transform is None:
                transform = adjust_transform_for_image(next(self.transform_generator), image, self.transform_parameters.relative_translation)

            # apply transformation to image
            image = apply_transform(transform, image, self.transform_parameters)

            # Transform the bounding boxes in the annotations.
            annotations['bboxes'] = annotations['bboxes'].copy()
            for index in range(annotations['bboxes'].shape[0]):
                annotations['bboxes'][index, :] = transform_aabb(transform, annotations['bboxes'][index, :])

        return image, annotations

    def random_transform_group(self, image_group, annotations_group):
        """ Randomly transforms each image and its annotations.
        """

        assert(len(image_group) == len(annotations_group))

        for index in range(len(image_group)):
            # transform a single group entry
            image_group[index], annotations_group[index] = self.random_transform_group_entry(image_group[index], annotations_group[index])

        return image_group, annotations_group

    def resize_image(self, image, min_side, max_side):
        """ Resize an image using image_min_side and image_max_side.
        """
        return resize_image(image, min_side=min_side, max_side=max_side)

    def resize_image_and_expand_dims(self, image):
        image, scale = self.resize_image(image, min_side=self.image_min_side, max_side=self.image_max_side)
        image = np.expand_dims(image, axis=0)
        return image, scale
        
    def split_image(self, image, thresh=0.8):

        images = []
        left_top_strides = []
        right_down_strides = []
        height, width = image.shape[0:2]
        if width >= height:
            x_stride = self.image_max_side
            y_stride = self.image_min_side
        else:
            x_stride = self.image_min_side
            y_stride = self.image_max_side

        def add_row_crop(start_y):
            start_x = 0
            while start_x + x_stride < width:
                images.append(image[start_y:start_y+y_stride, start_x:start_x+x_stride])
                left_top_strides.append([start_x, start_y])
                right_down_strides.append([start_x+x_stride, start_y+y_stride])
                start_x += int(x_stride * thresh)
            images.append(image[start_y:start_y+y_stride, width-x_stride:])
            left_top_strides.append([width-x_stride, start_y])
            right_down_strides.append([width, start_y+y_stride])

        start_y = 0
        while start_y + y_stride < height:
            add_row_crop(start_y)
            start_y += int(y_stride * thresh)
        add_row_crop(height-y_stride)
        
        return np.array(images), left_top_strides, right_down_strides
        
    def preprocess_group_entry(self, image, annotations):
        """ Preprocess image and its annotations.
        """
        # preprocess the image
        image = self.preprocess_image(image, mode=self.preprocess_image_mode)

        # wangtf: Try, if we do not resize image.
        # # resize image
        # image, image_scale = self.resize_image(image, min_side=self.image_min_side, max_side=self.image_max_side)

        # # apply resizing to annotations too
        # annotations['bboxes'] *= image_scale

        # convert to the wanted keras floatx
        image = keras.backend.cast_to_floatx(image)

        return image, annotations

    def preprocess_group(self, image_group, annotations_group):
        """ Preprocess each image and its annotations in its group.
        """
        assert(len(image_group) == len(annotations_group))

        for index in range(len(image_group)):
            # preprocess a single group entry
            image_group[index], annotations_group[index] = self.preprocess_group_entry(image_group[index], annotations_group[index])

        return image_group, annotations_group

    def group_images(self):
        """ Order the images according to self.order and makes groups of self.batch_size.
        """
        # determine the order of the images
        order = list(range(self.size()))
        if self.group_method == 'random':
            random.shuffle(order)
        elif self.group_method == 'ratio':
            order.sort(key=lambda x: self.image_aspect_ratio(x))

        # divide into groups, one group = one batch
        self.groups = [[order[x % len(order)] for x in range(i, i + self.batch_size)] for i in range(0, len(order), self.batch_size)]

    def compute_inputs(self, image_group):
        """ Compute inputs for the network using an image_group.
        """
        # get the max image shape
        max_shape = tuple(max(image.shape[x] for image in image_group) for x in range(3))

        # construct an image batch object
        image_batch = np.zeros((self.batch_size,) + max_shape, dtype=keras.backend.floatx())

        # copy all images to the upper left part of the image batch object
        for image_index, image in enumerate(image_group):
            image_batch[image_index, :image.shape[0], :image.shape[1], :image.shape[2]] = image

        if keras.backend.image_data_format() == 'channels_first':
            image_batch = image_batch.transpose((0, 3, 1, 2))

        return image_batch

    def generate_anchors(self, image_shape):
        anchor_params = None
        if self.config and 'anchor_parameters' in self.config:
            anchor_params = parse_anchor_parameters(self.config)
        return anchors_for_shape(image_shape, anchor_params=anchor_params, shapes_callback=self.compute_shapes)

    def compute_targets(self, image_group, annotations_group):
        """ Compute target outputs for the network using images and their annotations.
        """
        # get the max image shape
        max_shape = tuple(max(image.shape[x] for image in image_group) for x in range(3))
        anchors   = self.generate_anchors(max_shape)

        batches = self.compute_anchor_targets(
            anchors,
            image_group,
            annotations_group,
            self.num_classes()
        )

        return list(batches)

    def compute_input_output(self, group):
        """ Compute inputs and target outputs for the network.
        """
        # load images and annotations
        image_group       = self.load_image_group(group)
        annotations_group = self.load_annotations_group(group)

        # check validity of annotations
        image_group, annotations_group = self.filter_annotations(image_group, annotations_group, group)

        # randomly transform data
        image_group, annotations_group = self.random_transform_group(image_group, annotations_group)

        # perform preprocessing steps
        image_group, annotations_group = self.preprocess_group(image_group, annotations_group)

        # compute network inputs
        inputs = self.compute_inputs(image_group)

        # compute network targets
        targets = self.compute_targets(image_group, annotations_group)

        return inputs, targets

    def __len__(self):
        """
        Number of batches for generator.
        """

        return len(self.groups)

    def __getitem__(self, index):
        """
        Keras sequence method for generating batches.
        """
        group = self.groups[index]
        inputs, targets = self.compute_input_output(group)

        return inputs, targets
